/////////////////////////////////////////////////////////////////////////
// level functions

export class LevelTemplate {
    constructor(data) {

        // common variables
        this.count = BigInt(0)
        this.multiplier = BigInt(1)
        this.multiplierCost = BigInt(10)
        this.efficiency = BigInt(1)
        this.efficiencyCost = BigInt(1)
        this.unlocked = false
        this.unlockTime = 0

        // copy level specific data in to object
        Object.assign(this, data)
    }

    // save data
    save() {
        let save_data = { }
        let props = [ 
            'index', 'count', 
            'multiplier', 'multiplierCost',
            'efficiency', 'efficiencyCost',
            'unlocked', 'unlockTime'
        ]

        props.forEach((prop) => {
            save_data[prop] = this[prop]
        })
        return save_data
    }

    restore(save_data) {
        Object.assign(this, save_data)
    }

    // unlock this level
    unlock(elapsed) {
        this.unlocked = true
        this.unlockTime = elapsed
    }

    // upgrade functions
    canUpgrade() {
        return this.count >= this.multiplierCost
    }

    shouldAutobuy() {
        return this.count >= (this.multiplierCost * 2n)
    }

    upgradeMultiplier(state) {
        this.count -= this.multiplierCost
        this.multiplier *= 2n
        if (state.services.upgrade.level >=2) {
            for(let i = state.levels.length - 1; i >= 0; i--) {
                if (state.levels[i].unlocked) {
                    let multiplier = ((state.levels.length - i)*2) + 1
                    this.multiplierCost *= BigInt(multiplier)
                    break
                }
            }
        } else {
            this.multiplierCost *= 11n
        }
    }

    upgradeEfficiency(state) {
        // am I the last level ?
        if (this.index == (state.count - 1)) { return }

        let nextLevel = state.levels[this.index + 1]
        nextLevel.count -= this.efficiencyCost
        this.efficiency *= 2n
        this.efficiencyCost *= (state.services.upgrade.level >= 1) ? 7n : 13n
    }

    // calculate max unit cost as percentage of previous level units
    calcUnitCostPercent(state) {
        if (this.index == 0) { return 100 }

        let max = this.maxCanBuyInternal(state)
        let units = state.levels[this.index - 1].count
        if (units == 0) { return 100 }
        return Number((max * this.cost * 100n) / units)
    }

    // calculate max unit cost as percentage of available bandwidth
    calcBWCostPercent(state) {
        let max = this.maxCanBuyInternal(state)
        let bw = state.bandwidth.count
        if (bw == 0) { return 100 }
        return Number((max * 100n) / bw)
    }

    // unit purchasing
    buy(state, amount) {

        // take cost based on amount less the efficiency
        let a = amount / this.efficiency
        state.bandwidth.count -= a
        if (this.index != 0) {
            state.levels[this.index - 1].count -= a * this.cost
        }

        // increase units
        this.count += amount
    }

    // calculate the maximum number of units that can be bought
    maxCanBuyInternal(state) {
        let max

        if (this.index == 0) {
            // level zero is limited by bandwidth only
            max = state.bandwidth.count

        } else {
            // otherwise its the maximum of bw and previous level
            max = state.levels[this.index - 1].count / this.cost
            if (max > state.bandwidth.count) {
                max = state.bandwidth.count
            }
        }
        return max
    }

    maxCanBuy(state) {
        return this.maxCanBuyInternal(state) * this.efficiency
    }

    // timer tick
    tick(state, interval) {
        // am I the last level ?
        if (this.index == (state.count - 1)) { return }

        // add units to this level
        let nextLevel = state.levels[this.index + 1]
        this.count += nextLevel.count * nextLevel.multiplier

        // unlock next level if possible
        if ((this.count > 0) && (!nextLevel.unlocked)) { nextLevel.unlock(state.ui.elapsed) }
    }

}

/////////////////////////////////////////////////////////////////////////
// end of file