/////////////////////////////////////////////////////////////////////////

import { defineStore } from "pinia"
import { computed, reactive } from "vue"
import { LevelDefs } from './leveldefs.js'
import { Bandwidth } from './bandwidth.js'
import { Autobuyer } from "./autobuyer.js"
import { Services } from "./services.js"

/////////////////////////////////////////////////////////////////////////

export const useState = defineStore('state', () => {

    // timer tick
    function tick(state, i) {
        let interval = BigInt(i)
        ui.elapsed += i

        // update bandwidth
        bandwidth.tick(state, interval)

        // update levels
        levels.forEach((l) => {
            l.tick(state, interval)
        })

        ui.save = ((ui.elapsed % 30) == 0)
        if (ui.save) { save() }

        autobuyer.tick(state, i)
        services.tick(state, i)
    }

    // save and restore
    function reset() {
        localStorage.clear()
        location.reload()
    }

    function save() {

        let save_data = {
            version: version,
            elapsed: ui.elapsed,
            levels: Array.from(levels, (l) => l.save()),
            bandwidth: bandwidth.save(),
            services: services.save(),
            autobuyer: autobuyer.save()
        }

        // replacer func to handle big ints
        const replacer = (key, value) =>
            typeof value === 'bigint' ? '__big__' + value.toString() : value
        let encoded = JSON.stringify(save_data, replacer)

        localStorage.setItem('clicker42', encoded)
    }

    function restore() {
        let encoded = localStorage.getItem('clicker42')
        if (!encoded) { return }

        // reviver func to handle big ints
        const reviver = (key, value) =>
            ((typeof value === 'string') && (value.startsWith('__big__'))) ? BigInt(value.substring(7)) : value;
        let save_data = JSON.parse(encoded, reviver)

        // don't load older versions
        if (version > save_data.version) { return }
        if (typeof save_data.services.upgrade === 'undefined') {
            save_data.services.upgrade = {
                unlocked: false,
                level: 0
            }
        }

        ui.elapsed = save_data.elapsed

        // restore levels
        save_data.levels.forEach((l) => {
            let index = l.index
            levels[index].restore(l)
        })

        // and bandwidth
        bandwidth.restore(save_data.bandwidth)
        services.restore(save_data.services)
        autobuyer.restore(save_data.autobuyer)
    }

    function noticeClear() {
        ui.notice = '<span class="p-2">&nbsp;</span>'
    }

    function noticeTemplate(style, text) {
        ui.notice = '<span class="badge p-2 rounded-pill ' + style + '">' + text + '</span>'
    }

    function noticeSuccess(text) { noticeTemplate('bg-success', text) }
    function noticeDanger(text) { noticeTemplate('bg-danger', text) }

    // game data
    const version = 0.1
    const ui = reactive({
        selectedLevel: 0,
        elapsed: 0,
        notice: '<span class="p-2">&nbsp;</span>',
        save: false
    })
    const autobuyer = reactive(new Autobuyer())

    // level data
    const levels = reactive(LevelDefs)
    levels.forEach((l, ix) => { 
        l.index = ix
        l.multiplier = BigInt(ix)
    })

    const bandwidth = reactive(new Bandwidth())
    const services = reactive(new Services())

    // computed functions
    const count = computed(() => levels.length)

    // return the things that will form part of the store
    return {
        // data
        version, ui, levels, bandwidth, services, autobuyer,
        // computed
        count,
        // functions
        tick, reset, save, restore,
        noticeClear, noticeSuccess, noticeDanger
    }

})

/////////////////////////////////////////////////////////////////////////
// end of file